"""
for f in /tmp/Dropbox/work/odb_server_bundle/server/src/odbsvr/test/dhshome/data/root/20151108/12/16030.odb \
    /tmp/Dropbox/work/odb_server_bundle/server/src/odbsvr/test/dhshome/data/root/20151108/12/16029.odb \
    /tmp/Dropbox/work/odb_server_bundle/server/src/odbsvr/test/dhshome/data/root/20151108/12/16045.odb \
    /tmp/Dropbox/work/odb_server_bundle/server/src/odbsvr/test/dhshome/data/root/20151108/12/16065.odb \
     /tmp/Dropbox/work/odb_server_bundle/server/src/odbsvr/test/dhshome/data/root/20151108/12/16009.odb ; do
     odb count $f
 done
 """
 python -c 'print 119256 + 15151 + 31144 + 18800 + 93'

