// Classes user by swig

#include "eckit/eckit.h"

#include "eckit/sql/SQLParser.h"
#include "odc/DispatchingWriter.h"
#include "odc/IteratorProxy.h"
#include "odc/ODBAPISettings.h"
#include "odc/ODBAPIVersion.h"
#include "odc/Reader.h"
#include "odc/Select.h"
#include "odc/WriterBufferingIterator.h"
#include "odc/core/Column.h"
#include "odc/core/MetaData.h"
#include "odc/csv/TextReader.h"
#include "odc/csv/TextReaderIterator.h"

using namespace odc;
